@@ -4,7 +4,7 @@ use gimli::{EndianReader, LittleEndian};
 use crate::arcdata::ArcData;
 use crate::arch::Arch;
 use crate::cache::{AllocationPolicy, Cache};
-use crate::dwarf::{DwarfUnwinder, DwarfUnwinding};
+use crate::dwarf::{DwarfCfiIndex, DwarfUnwinder, DwarfUnwinding, UnwindSectionType};
 use crate::error::{Error, UnwinderError};
 use crate::instruction_analysis::InstructionAnalysis;
 use crate::macho::{
@@ -329,11 +329,12 @@ impl<
         F: FnMut(u64) -> Result<u64, ()>,
     {
         let unwind_result = match &module.unwind_data {
-            ModuleUnwindData::CompactUnwindInfoAndEhFrame(unwind_data, eh_frame_data) => {
+            ModuleUnwindDataInternal::CompactUnwindInfoAndEhFrame(unwind_data, eh_frame_data) => {
                 // eprintln!("unwinding with cui and eh_frame in module {}", module.name);
                 let text_bytes = module.text_data.as_ref().and_then(|data| {
                     let offset_from_base =
-                        u32::try_from(data.range.start.checked_sub(module.base_address)?).ok()?;
+                        u32::try_from(data.address_range.start.checked_sub(module.base_address)?)
+                            .ok()?;
                     Some(TextBytes::new(offset_from_base, &data.bytes[..]))
                 });
                 let stubs_range = if let Some(stubs_range) = &module.sections.stubs {
@@ -371,6 +372,7 @@ impl<
                         };
                         let mut dwarf_unwinder = DwarfUnwinder::<_, A, P::GimliStorage>::new(
                             EndianReader::new(eh_frame_data, LittleEndian),
+                            UnwindSectionType::EhFrame,
                             None,
                             &mut cache.eh_frame_unwind_context,
                             &module.sections,
@@ -380,12 +382,13 @@ impl<
                     }
                 }
             }
-            ModuleUnwindData::EhFrameHdrAndEhFrame(eh_frame_hdr, eh_frame_data) => {
-                let eh_frame_hdr_data = ArcData(eh_frame_hdr.clone());
+            ModuleUnwindDataInternal::EhFrameHdrAndEhFrame(eh_frame_hdr, eh_frame_data) => {
+                let eh_frame_hdr_data = &eh_frame_hdr[..];
                 let eh_frame_data = ArcData(eh_frame_data.clone());
                 let mut dwarf_unwinder = DwarfUnwinder::<_, A, P::GimliStorage>::new(
                     EndianReader::new(eh_frame_data, LittleEndian),
-                    Some(EndianReader::new(eh_frame_hdr_data, LittleEndian)),
+                    UnwindSectionType::EhFrame,
+                    Some(eh_frame_hdr_data),
                     &mut cache.eh_frame_unwind_context,
                     &module.sections,
                 );
@@ -394,30 +397,114 @@ impl<
                     .ok_or(UnwinderError::EhFrameHdrCouldNotFindAddress)?;
                 dwarf_unwinder.unwind_frame_with_fde(regs, address, fde_offset, read_stack)?
             }
-            ModuleUnwindData::EhFrame(_) => {
-                return Err(UnwinderError::UnhandledModuleUnwindDataType)
+            ModuleUnwindDataInternal::DwarfCfiIndexAndEhFrame(index, eh_frame_data) => {
+                let eh_frame_data = ArcData(eh_frame_data.clone());
+                let mut dwarf_unwinder = DwarfUnwinder::<_, A, P::GimliStorage>::new(
+                    EndianReader::new(eh_frame_data, LittleEndian),
+                    UnwindSectionType::EhFrame,
+                    None,
+                    &mut cache.eh_frame_unwind_context,
+                    &module.sections,
+                );
+                let fde_offset = index
+                    .fde_offset_for_relative_address(rel_lookup_address)
+                    .ok_or(UnwinderError::EhFrameHdrCouldNotFindAddress)?;
+                dwarf_unwinder.unwind_frame_with_fde(regs, address, fde_offset, read_stack)?
             }
-            ModuleUnwindData::None => return Err(UnwinderError::NoModuleUnwindData),
+            ModuleUnwindDataInternal::DwarfCfiIndexAndDebugFrame(index, debug_frame_data) => {
+                let debug_frame_data = ArcData(debug_frame_data.clone());
+                let mut dwarf_unwinder = DwarfUnwinder::<_, A, P::GimliStorage>::new(
+                    EndianReader::new(debug_frame_data, LittleEndian),
+                    UnwindSectionType::DebugFrame,
+                    None,
+                    &mut cache.eh_frame_unwind_context,
+                    &module.sections,
+                );
+                let fde_offset = index
+                    .fde_offset_for_relative_address(rel_lookup_address)
+                    .ok_or(UnwinderError::EhFrameHdrCouldNotFindAddress)?;
+                dbg!(dwarf_unwinder.unwind_frame_with_fde(regs, address, fde_offset, read_stack))?
+            }
+            ModuleUnwindDataInternal::None => return Err(UnwinderError::NoModuleUnwindData),
         };
         Ok(unwind_result)
     }
 }
 
 pub enum ModuleUnwindData<D: Deref<Target = [u8]>> {
-    CompactUnwindInfoAndEhFrame(D, Option<Arc<D>>),
-    EhFrameHdrAndEhFrame(Arc<D>, Arc<D>),
-    EhFrame(Arc<D>),
+    CompactUnwindInfoAndEhFrame(D, Option<D>),
+    EhFrameHdrAndEhFrame(D, D),
+    EhFrame(D),
+    DebugFrame(D),
     None,
 }
 
+enum ModuleUnwindDataInternal<D: Deref<Target = [u8]>> {
+    CompactUnwindInfoAndEhFrame(D, Option<Arc<D>>),
+    EhFrameHdrAndEhFrame(D, Arc<D>),
+    DwarfCfiIndexAndEhFrame(DwarfCfiIndex, Arc<D>),
+    DwarfCfiIndexAndDebugFrame(DwarfCfiIndex, Arc<D>),
+    None,
+}
+
+impl<D: Deref<Target = [u8]>> ModuleUnwindDataInternal<D> {
+    fn new(
+        unwind_data: ModuleUnwindData<D>,
+        sections: &ModuleSectionAddressRanges,
+        base_address: u64,
+    ) -> Self {
+        match unwind_data {
+            ModuleUnwindData::CompactUnwindInfoAndEhFrame(cui, eh_frame) => {
+                ModuleUnwindDataInternal::CompactUnwindInfoAndEhFrame(cui, eh_frame.map(Arc::new))
+            }
+            ModuleUnwindData::EhFrameHdrAndEhFrame(eh_frame_hdr, eh_frame) => {
+                ModuleUnwindDataInternal::EhFrameHdrAndEhFrame(eh_frame_hdr, Arc::new(eh_frame))
+            }
+            ModuleUnwindData::EhFrame(eh_frame) => {
+                match DwarfCfiIndex::try_new_eh_frame(&eh_frame, sections, base_address) {
+                    Ok(index) => {
+                        ModuleUnwindDataInternal::DwarfCfiIndexAndEhFrame(index, Arc::new(eh_frame))
+                    }
+                    Err(_) => ModuleUnwindDataInternal::None,
+                }
+            }
+            ModuleUnwindData::DebugFrame(debug_frame) => {
+                match DwarfCfiIndex::try_new_debug_frame(&debug_frame, sections, base_address) {
+                    Ok(index) => ModuleUnwindDataInternal::DwarfCfiIndexAndDebugFrame(
+                        index,
+                        Arc::new(debug_frame),
+                    ),
+                    Err(_) => ModuleUnwindDataInternal::None,
+                }
+            }
+            ModuleUnwindData::None => ModuleUnwindDataInternal::None,
+        }
+    }
+}
+
 pub struct TextByteData<D: Deref<Target = [u8]>> {
     bytes: D,
-    range: Range<u64>,
+    address_range: Range<u64>,
 }
 
 impl<D: Deref<Target = [u8]>> TextByteData<D> {
-    pub fn new(bytes: D, range: Range<u64>) -> Self {
-        Self { bytes, range }
+    pub fn new(bytes: D, address_range: Range<u64>) -> Self {
+        Self {
+            bytes,
+            address_range,
+        }
+    }
+
+    pub fn get_bytes(&self, address_range: Range<u64>) -> Option<&[u8]> {
+        let rel_start = address_range.start.checked_sub(self.address_range.start)?;
+        let rel_start = usize::try_from(rel_start).ok()?;
+        let rel_end = address_range.end.checked_sub(self.address_range.start)?;
+        let rel_end = usize::try_from(rel_end).ok()?;
+        self.bytes.get(rel_start..rel_end)
+    }
+
+    pub fn address_range(&self) -> Range<u64> {
+        self.address_range.clone()
     }
 }
 
@@ -427,7 +514,7 @@ pub struct Module<D: Deref<Target = [u8]>> {
     address_range: Range<u64>,
     base_address: u64,
     sections: ModuleSectionAddressRanges,
-    unwind_data: ModuleUnwindData<D>,
+    unwind_data: ModuleUnwindDataInternal<D>,
     text_data: Option<TextByteData<D>>,
 }
 
@@ -451,6 +538,7 @@ impl<D: Deref<Target = [u8]>> Module<D> {
         unwind_data: ModuleUnwindData<D>,
         text_data: Option<TextByteData<D>>,
     ) -> Self {
+        let unwind_data = ModuleUnwindDataInternal::new(unwind_data, &sections, base_address);
         Self {
             name,
             address_range,
